/*
 * Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
 * license agreements.  See the NOTICE file distributed with this work for
 * additional information regarding copyright ownership.  Crate licenses
 * this file to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * However, if you have executed another commercial license agreement
 * with Crate these terms will supersede the license and you may use the
 * software solely pursuant to the terms of the relevant commercial agreement.
 */

package io.crate.executor.transport;

import com.google.common.base.Function;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.SettableFuture;
import io.crate.action.job.*;
import io.crate.core.collections.Bucket;
import io.crate.executor.JobTask;
import io.crate.executor.TaskResult;
import io.crate.jobs.*;
import io.crate.operation.*;
import io.crate.operation.projectors.RowReceiver;
import io.crate.planner.node.ExecutionPhase;
import io.crate.planner.node.ExecutionPhases;
import io.crate.planner.node.NodeOperationGrouper;
import org.elasticsearch.action.ActionListener;
import org.elasticsearch.cluster.ClusterService;
import org.elasticsearch.common.collect.Tuple;
import org.elasticsearch.common.logging.ESLogger;
import org.elasticsearch.common.logging.Loggers;
import org.elasticsearch.indices.IndicesService;

import javax.annotation.Nullable;
import java.util.*;


public class ExecutionPhasesTask extends JobTask {

    private static final ESLogger LOGGER = Loggers.getLogger(ExecutionPhasesTask.class);

    private final TransportJobAction transportJobAction;
    private final List<NodeOperationTree> nodeOperationTrees;
    private final OperationType operationType;
    private final ClusterService clusterService;
    private ContextPreparer contextPreparer;
    private final JobContextService jobContextService;
    private final IndicesService indicesService;


    private final List<SettableFuture<TaskResult>> results = new ArrayList<>();
    private boolean hasDirectResponse;

    public enum OperationType {
        BULK,
        /**
         * UNKNOWN means it will depend on the number of nodeOperationTrees, if > 1 it will use bulk, otherwise QueryResult
         */
        UNKNOWN
    }

    protected ExecutionPhasesTask(UUID jobId,
                                  ClusterService clusterService,
                                  ContextPreparer contextPreparer,
                                  JobContextService jobContextService,
                                  IndicesService indicesService,
                                  TransportJobAction transportJobAction,
                                  List<NodeOperationTree> nodeOperationTrees,
                                  OperationType operationType) {
        super(jobId);
        this.clusterService = clusterService;
        this.contextPreparer = contextPreparer;
        this.jobContextService = jobContextService;
        this.indicesService = indicesService;
        this.transportJobAction = transportJobAction;
        this.nodeOperationTrees = nodeOperationTrees;
        this.operationType = operationType;

        for (NodeOperationTree nodeOperationTree : nodeOperationTrees) {
            results.add(SettableFuture.<TaskResult>create());
            for (NodeOperation nodeOperation : nodeOperationTree.nodeOperations()) {
                if (ExecutionPhases.hasDirectResponseDownstream(nodeOperation.downstreamNodes())) {
                    hasDirectResponse = true;
                    break;
                }
            }
        }
    }

    @Override
    public void start() {
        FluentIterable<NodeOperation> nodeOperations = FluentIterable.from(nodeOperationTrees)
                .transformAndConcat(new Function<NodeOperationTree, Iterable<? extends NodeOperation>>() {
                    @Nullable
                    @Override
                    public Iterable<? extends NodeOperation> apply(NodeOperationTree input) {
                        return input.nodeOperations();
                    }
                });

        Map<String, Collection<NodeOperation>> operationByServer = NodeOperationGrouper.groupByServer(nodeOperations);
        List<PageDownstreamContext> pageDownstreamContexts = new ArrayList<>(nodeOperationTrees.size());
        List<Tuple<ExecutionPhase, RowReceiver>> handlerPhases = new ArrayList<>(nodeOperationTrees.size());

        if (operationType == OperationType.BULK || nodeOperationTrees.size() > 1) {
            // bulk Operation with rowCountResult
            for (int i = 0; i < nodeOperationTrees.size(); i++) {
                SettableFuture<TaskResult> result = results.get(i);
                RowCountResultRowDownstream rowDownstream = new RowCountResultRowDownstream(result);
                handlerPhases.add(new Tuple<ExecutionPhase, RowReceiver>(nodeOperationTrees.get(i).leaf(), rowDownstream));
            }
        } else {
            SettableFuture<TaskResult> result = Iterables.getOnlyElement(results);
            QueryResultRowDownstream downstream = new QueryResultRowDownstream(result);
            handlerPhases.add(new Tuple<ExecutionPhase, RowReceiver>(Iterables.getOnlyElement(nodeOperationTrees).leaf(), downstream));
        }

        try {
            setupContext(operationByServer, pageDownstreamContexts, handlerPhases);
        } catch (Throwable throwable) {
            for (SettableFuture<TaskResult> result : results) {
                result.setException(throwable);
            }
        }
        if (operationByServer.isEmpty()) {
            return;
        }
        sendJobRequests(pageDownstreamContexts, operationByServer);
    }

    private void setupContext(Map<String, Collection<NodeOperation>> operationByServer,
                              List<PageDownstreamContext> pageDownstreamContexts,
                              List<Tuple<ExecutionPhase, RowReceiver>> handlerPhases) throws Throwable {
        boolean noNodeOperations = false;
        if (operationByServer.isEmpty()) {
            // this should never happen, instead we should use a NOOP plan instead
            // which immediately results in an empty result
            noNodeOperations = true;
        }
        List<ExecutionSubContext> localContextAndStartOperation = createLocalContextAndStartOperation(operationByServer, handlerPhases);
        if (noNodeOperations) {
            for (ExecutionSubContext executionSubContext : localContextAndStartOperation) {
                executionSubContext.close();
            }
            return;
        }
        if (hasDirectResponse) {
            for (ExecutionSubContext executionSubContext : localContextAndStartOperation) {
                assert executionSubContext != null && executionSubContext instanceof DownstreamExecutionSubContext
                        : "Need DownstreamExecutionSubContext for DIRECT_RESPONSE remote jobs";
                pageDownstreamContexts.add(((DownstreamExecutionSubContext) executionSubContext).pageDownstreamContext((byte) 0));
            }
        }
    }

    private void sendJobRequests(List<PageDownstreamContext> pageDownstreamContexts,
                                 Map<String, Collection<NodeOperation>> operationsByServer) {
        int idx = 0;
        for (Map.Entry<String, Collection<NodeOperation>> entry : operationsByServer.entrySet()) {
            String serverNodeId = entry.getKey();
            Collection<NodeOperation> nodeOperations = entry.getValue();

            JobRequest request = new JobRequest(jobId(), nodeOperations);
            if (hasDirectResponse) {
                transportJobAction.execute(serverNodeId, request, new DirectResponseListener(idx, pageDownstreamContexts));
            } else {
                transportJobAction.execute(serverNodeId, request,
                        new FailureOnlyResponseListener(results));
            }
            idx++;
        }
    }

    /**
     * removes the localNodeId entry from the nodesByServer map and initializes the context and starts the operation.
     *
     * This is done in order to be able to create the JobExecutionContext with the localMerge PageDownstreamContext
     */
    private List<ExecutionSubContext> createLocalContextAndStartOperation(Map<String, Collection<NodeOperation>> operationsByServer,
                                                                          List<Tuple<ExecutionPhase, RowReceiver>> handlerPhases) throws Throwable {
        String localNodeId = clusterService.localNode().id();
        JobExecutionContext.Builder builder = jobContextService.newBuilder(jobId());
        Collection<NodeOperation> localNodeOperations = null;
        SharedShardContexts sharedShardContexts = null;
        if (!hasDirectResponse) {
            localNodeOperations = operationsByServer.remove(localNodeId);
        }
        if (localNodeOperations == null) {
            localNodeOperations = ImmutableList.of();
        } else {
            sharedShardContexts = new SharedShardContexts(indicesService);
        }

        List<ExecutionSubContext> executionSubContexts = contextPreparer.prepareOnHandler(
                jobId(),
                localNodeOperations,
                builder,
                handlerPhases,
                sharedShardContexts
        );

        if (!hasDirectResponse) {
            JobExecutionContext context = jobContextService.createContext(builder);
            context.start();
        } else {
            for (ExecutionSubContext executionSubContext : executionSubContexts) {
                executionSubContext.prepare();
                executionSubContext.start();
            }
        }
        return executionSubContexts;
    }

    @Override
    public List<? extends ListenableFuture<TaskResult>> result() {
        return results;
    }

    @Override
    public void upstreamResult(List<? extends ListenableFuture<TaskResult>> result) {
        throw new UnsupportedOperationException("ExecutionNodesTask doesn't support upstreamResult");
    }

    private static class DirectResponseListener implements ActionListener<JobResponse> {

        private final int bucketIdx;
        private final List<PageDownstreamContext> pageDownstreamContexts;

        public DirectResponseListener(int bucketIdx, List<PageDownstreamContext> pageDownstreamContexts) {
            this.bucketIdx = bucketIdx;
            this.pageDownstreamContexts = pageDownstreamContexts;
        }

        @Override
        public void onResponse(JobResponse jobResponse) {
            for (int i = 0; i < pageDownstreamContexts.size(); i++) {
                PageDownstreamContext pageDownstreamContext = pageDownstreamContexts.get(i);
                jobResponse.streamers(pageDownstreamContext.streamer());
                Bucket bucket = jobResponse.directResponse().get(i);
                if (bucket == null) {
                    pageDownstreamContext.failure(bucketIdx, new IllegalStateException("expected directResponse but didn't get one"));
                }
                pageDownstreamContext.setBucket(bucketIdx, bucket, true, new PageResultListener() {
                    @Override
                    public void needMore(boolean needMore) {
                        if (needMore) {
                            LOGGER.warn("requested more data but directResponse doesn't support paging");
                        }
                    }

                    @Override
                    public int buckedIdx() {
                        return bucketIdx;
                    }
                });
            }
        }

        @Override
        public void onFailure(Throwable e) {
            for (PageDownstreamContext pageDownstreamContext : pageDownstreamContexts) {
                pageDownstreamContext.failure(bucketIdx, e);
            }
        }
    }

    private static class FailureOnlyResponseListener implements ActionListener<JobResponse> {

        private final List<SettableFuture<TaskResult>> results;

        public FailureOnlyResponseListener(List<SettableFuture<TaskResult>> results) {
            this.results = results;
        }

        @Override
        public void onResponse(JobResponse jobResponse) {
            if (jobResponse.directResponse().size() > 0) {
                for (SettableFuture<TaskResult> result : results) {
                    result.setException(new IllegalStateException("Got a directResponse but didn't expect one"));
                }
            }
        }

        @Override
        public void onFailure(Throwable e) {
            // in the non-direct-response case the failure is pushed to downStreams
            LOGGER.warn(e.getMessage(), e);
        }
    }
}
